@@ -3,6 +3,10 @@
  * Copyright (C) 2001-2003, Ximian, Inc.
  */
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -12,29 +16,33 @@
 #include <sys/stat.h>
 #include <unistd.h>
 
+#ifdef HAVE_GNOME
+#include <libsoup/soup-gnome.h>
+#else
 #include <libsoup/soup.h>
+#endif
 
 #ifdef G_OS_WIN32
 #include <io.h>
 #define mkdir(path, mode) _mkdir (path)
 #endif
 
-SoupSession *session;
-GMainLoop *loop;
-gboolean recurse = FALSE, debug = FALSE;
-const char *method = SOUP_METHOD_GET;
-char *base;
-SoupUri *base_uri;
-int pending;
+static SoupSession *session;
+static GMainLoop *loop;
+static gboolean recurse = FALSE, debug = FALSE;
+static const char *method;
+static char *base;
+static SoupURI *base_uri;
+static GHashTable *fetched_urls;
 
 static GPtrArray *
-find_hrefs (const SoupUri *base, const char *body, int length)
+find_hrefs (SoupURI *doc_base, const char *body, int length)
 {
 	GPtrArray *hrefs = g_ptr_array_new ();
 	char *buf = g_strndup (body, length);
 	char *start = buf, *end;
 	char *href, *frag;
-	SoupUri *uri;
+	SoupURI *uri;
 
 	while ((start = strstr (start, "href"))) {
 		start += 4;
@@ -57,19 +65,19 @@ find_hrefs (const SoupUri *base, const char *body, int length)
 		if (frag)
 			*frag = '\0';
 
-		uri = soup_uri_new_with_base (base, href);
+		uri = soup_uri_new_with_base (doc_base, href);
 		g_free (href);
 
 		if (!uri)
 			continue;
-		if (base->protocol != uri->protocol ||
-		    base->port != uri->port ||
-		    g_ascii_strcasecmp (base->host, uri->host) != 0) {
+		if (doc_base->scheme != uri->scheme ||
+		    doc_base->port != uri->port ||
+		    g_ascii_strcasecmp (doc_base->host, uri->host) != 0) {
 			soup_uri_free (uri);
 			continue;
 		}
 
-		if (strncmp (base->path, uri->path, strlen (base->path)) != 0) {
+		if (strncmp (doc_base->path, uri->path, strlen (doc_base->path)) != 0) {
 			soup_uri_free (uri);
 			continue;
 		}
@@ -98,92 +106,20 @@ mkdirs (const char *path)
 	}
 }
 
-static void get_url (const char *url);
-
-static void
-print_header (gpointer name, gpointer value, gpointer data)
-{
-	printf ("%s: %s\n", (const char *)name, (const char *)value);
-}
-
-static void
-got_url (SoupMessage *msg, gpointer uri)
-{
-	char *name;
-	int fd, i;
-	GPtrArray *hrefs;
-	const char *header;
-
-	name = soup_message_get_uri (msg)->path;
-	if (strncmp (base_uri->path, name, strlen (base_uri->path)) != 0) {
-		fprintf (stderr, "  Error: not under %s\n", base_uri->path);
-		goto DONE;
-	}
-	if (debug) {
-		char *path = soup_uri_to_string (soup_message_get_uri (msg), TRUE);
-		printf ("%s %s HTTP/1.%d\n\n", method, path,
-			soup_message_get_http_version (msg));
-		printf ("HTTP/1.%d %d %s\n",
-			soup_message_get_http_version (msg),
-			msg->status_code, msg->reason_phrase);
-		soup_message_foreach_header (msg->response_headers, print_header, NULL);
-		printf ("\n");
-	} else
-		printf ("%s: %d %s\n", name, msg->status_code, msg->reason_phrase);
-
-	name += strlen (base_uri->path);
-	if (*name == '/')
-		name++;
-
-	if (SOUP_STATUS_IS_REDIRECTION (msg->status_code)) {
-		unlink (name);
-		header = soup_message_get_header (msg->response_headers, "Location");
-		if (header) {
-			if (!debug)
-				printf ("  -> %s\n", header);
-			get_url (header);
-		}
-		goto DONE;
-	}
-
-	if (!SOUP_STATUS_IS_SUCCESSFUL (msg->status_code))
-		goto DONE;
-
-	if (recurse)
-		fd = open (name, O_WRONLY | O_CREAT | O_TRUNC, 0644);
-	else
-		fd = STDOUT_FILENO;
-	write (fd, msg->response.body, msg->response.length);
-	if (!recurse)
-		goto DONE;
-	close (fd);
-
-	header = soup_message_get_header (msg->response_headers, "Content-Type");
-	if (header && g_ascii_strncasecmp (header, "text/html", 9) != 0)
-		goto DONE;
-
-	hrefs = find_hrefs (uri, msg->response.body, msg->response.length);
-	for (i = 0; i < hrefs->len; i++) {
-		get_url (hrefs->pdata[i]);
-		g_free (hrefs->pdata[i]);
-	}
-	g_ptr_array_free (hrefs, TRUE);
-
- DONE:
-	soup_uri_free (uri);
-	if (!--pending)
-		g_main_quit (loop);
-}
-
 static void
 get_url (const char *url)
 {
 	char *url_to_get, *slash, *name;
 	SoupMessage *msg;
-	int fd;
+	int fd, i;
+	SoupURI *uri;
+	GPtrArray *hrefs;
+	const char *header;
 
 	if (strncmp (url, base, strlen (base)) != 0)
 		return;
+	if (strchr (url, '?') && strcmp (url, base) != 0)
+		return;
 
 	slash = strrchr (url, '/');
 	if (slash && !slash[1])
@@ -191,6 +127,10 @@ get_url (const char *url)
 	else
 		url_to_get = g_strdup (url);
 
+	if (g_hash_table_lookup (fetched_urls, url_to_get))
+		return;
+	g_hash_table_insert (fetched_urls, url_to_get, url_to_get);
+
 	if (recurse) {
 		/* See if we're already downloading it, and create the
 		 * file if not.
@@ -199,10 +139,8 @@ get_url (const char *url)
 		name = url_to_get + strlen (base);
 		if (*name == '/')
 			name++;
-		if (access (name, F_OK) == 0) {
-			g_free (url_to_get);
+		if (access (name, F_OK) == 0)
 			return;
-		}
 
 		mkdirs (name);
 		fd = open (name, O_WRONLY | O_CREAT | O_TRUNC, 0644);
@@ -212,9 +150,73 @@ get_url (const char *url)
 	msg = soup_message_new (method, url_to_get);
 	soup_message_set_flags (msg, SOUP_MESSAGE_NO_REDIRECT);
 
-	pending++;
-	soup_session_queue_message (session, msg, got_url, soup_uri_new (url));
-	g_free (url_to_get);
+	soup_session_send_message (session, msg);
+
+	name = soup_message_get_uri (msg)->path;
+	if (strncmp (base_uri->path, name, strlen (base_uri->path)) != 0) {
+		fprintf (stderr, "  Error: not under %s\n", base_uri->path);
+		return;
+	}
+
+	if (debug) {
+		SoupMessageHeadersIter iter;
+		const char *hname, *value;
+		char *path = soup_uri_to_string (soup_message_get_uri (msg), TRUE);
+
+		printf ("%s %s HTTP/1.%d\n\n", method, path,
+			soup_message_get_http_version (msg));
+		printf ("HTTP/1.%d %d %s\n",
+			soup_message_get_http_version (msg),
+			msg->status_code, msg->reason_phrase);
+
+		soup_message_headers_iter_init (&iter, msg->response_headers);
+		while (soup_message_headers_iter_next (&iter, &hname, &value))
+			printf ("%s: %s\r\n", hname, value);
+		printf ("\n");
+	} else
+		printf ("%s: %d %s\n", name, msg->status_code, msg->reason_phrase);
+
+	name += strlen (base_uri->path);
+	if (*name == '/')
+		name++;
+
+	if (SOUP_STATUS_IS_REDIRECTION (msg->status_code)) {
+		if (recurse)
+			unlink (name);
+		header = soup_message_headers_get_one (msg->response_headers,
+						       "Location");
+		if (header) {
+			if (!debug)
+				printf ("  -> %s\n", header);
+			get_url (header);
+		}
+		return;
+	}
+
+	if (!SOUP_STATUS_IS_SUCCESSFUL (msg->status_code))
+		return;
+
+	if (recurse)
+		fd = open (name, O_WRONLY | O_CREAT | O_TRUNC, 0644);
+	else
+		fd = STDOUT_FILENO;
+	write (fd, msg->response_body->data, msg->response_body->length);
+	if (!recurse)
+		return;
+	close (fd);
+
+	header = soup_message_headers_get_content_type (msg->response_headers, NULL);
+	if (header && g_ascii_strcasecmp (header, "text/html") != 0)
+		return;
+
+	uri = soup_uri_new (url);
+	hrefs = find_hrefs (uri, msg->response_body->data, msg->response_body->length);
+	soup_uri_free (uri);
+	for (i = 0; i < hrefs->len; i++) {
+		get_url (hrefs->pdata[i]);
+		g_free (hrefs->pdata[i]);
+	}
+	g_ptr_array_free (hrefs, TRUE);
 }
 
 static void
@@ -228,13 +230,16 @@ int
 main (int argc, char **argv)
 {
 	const char *cafile = NULL;
-	SoupUri *proxy = NULL;
+	SoupURI *proxy = NULL;
+	gboolean synchronous = FALSE;
 	int opt;
 
-	g_type_init ();
 	g_thread_init (NULL);
+	g_type_init ();
 
-	while ((opt = getopt (argc, argv, "c:dhp:r")) != -1) {
+	method = SOUP_METHOD_GET;
+
+	while ((opt = getopt (argc, argv, "c:dhp:rs")) != -1) {
 		switch (opt) {
 		case 'c':
 			cafile = optarg;
@@ -262,6 +267,10 @@ main (int argc, char **argv)
 			recurse = TRUE;
 			break;
 
+		case 's':
+			synchronous = TRUE;
+			break;
+
 		case '?':
 			usage ();
 			break;
@@ -279,10 +288,35 @@ main (int argc, char **argv)
 		exit (1);
 	}
 
-	session = soup_session_async_new_with_options (
-		SOUP_SESSION_SSL_CA_FILE, cafile,
-		SOUP_SESSION_PROXY_URI, proxy,
-		NULL);
+	fetched_urls = g_hash_table_new (g_str_hash, g_str_equal);
+
+	if (synchronous) {
+		session = soup_session_sync_new_with_options (
+			SOUP_SESSION_SSL_CA_FILE, cafile,
+#ifdef HAVE_GNOME
+			SOUP_SESSION_ADD_FEATURE_BY_TYPE, SOUP_TYPE_GNOME_FEATURES_2_26,
+#endif
+			SOUP_SESSION_USER_AGENT, "get ",
+			NULL);
+	} else {
+		session = soup_session_async_new_with_options (
+			SOUP_SESSION_SSL_CA_FILE, cafile,
+#ifdef HAVE_GNOME
+			SOUP_SESSION_ADD_FEATURE_BY_TYPE, SOUP_TYPE_GNOME_FEATURES_2_26,
+#endif
+			SOUP_SESSION_USER_AGENT, "get ",
+			NULL);
+	}
+
+	/* Need to do this after creating the session, since adding
+	 * SOUP_TYPE_GNOME_FEATURE_2_26 will add a proxy resolver, thereby
+	 * bashing over the manually-set proxy.
+	 */
+	if (proxy) {
+		g_object_set (G_OBJECT (session), 
+			      SOUP_SESSION_PROXY_URI, proxy,
+			      NULL);
+	}
 
 	if (recurse) {
 		char *outdir;
@@ -297,11 +331,13 @@ main (int argc, char **argv)
 		g_free (outdir);
 	}
 
+	if (!synchronous)
+		loop = g_main_loop_new (NULL, TRUE);
+
 	get_url (base);
 
-	loop = g_main_loop_new (NULL, TRUE);
-	g_main_run (loop);
-	g_main_loop_unref (loop);
+	if (!synchronous)
+		g_main_loop_unref (loop);
 
 	soup_uri_free (base_uri);
 
